﻿using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using UnityEngine.Events;
using System;
using System.Linq;
using System.Text;
using System.Net;
using System.Security.Cryptography;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using AssetBundles;
using SimpleJSON;
using Random2 = System.Random;
using Random = UnityEngine.Random;

namespace SPQR
{
    public static class SPQRUtils
    {

        public static string EncodeJson_p67(string s)
        {

            s = s.Replace("\\'", "'");
            s = s.Replace("'", "\\'");
            return s;
        }
        public static string GetHttpPrefix()
        {
            //@todo fix obfuscation // comment replacement
            return "http:/" + "/";
        }

        public static string HashDecode_p67(string s)
        {
            string r = s.Replace(".", "+");
            r = r.Replace("_", "/");
            r = r.Replace("-", "=");

            byte[] data = Convert.FromBase64String(r);
            string decodedString = Encoding.UTF8.GetString(data);
            return decodedString;
        }

        public static string HashEncode_p67(string s)
        {

            var plainTextBytes = System.Text.Encoding.UTF8.GetBytes(s);
            string encoded = System.Convert.ToBase64String(plainTextBytes);
            string r = encoded.Replace("+", ".");
            r = r.Replace("/", "_");
            r = r.Replace("=", "-");

            return r;
        }

        public static void PrintComponents_p67(Atom A)
        {
            var cs = A.GetComponentsInChildren<Transform>();
            foreach (var c in cs)
            {
                SuperController.LogMessage("comp: " + c.name + " - " + c.GetType());
            }

        }

        public static MVRScript GetPlugin_p67(Atom A, string plugin)
        {
            var plugins = A.GetComponentsInChildren<MVRScript>();
            foreach (var c in plugins)
            {
                if (c.GetType().ToString() == plugin) return c;
            }

            return null;
        }

        public static string GetFilename_p67(string s)
        {
            int lastIndex = s.LastIndexOf('/');
            var name = s.Substring(0, lastIndex);
            var ext = s.Substring(lastIndex + 1);
            ext = ext.Replace("/", "");
            ext = ext.Replace("\\", "");
            return ext;
        }
        public static HSVColor GetColorByHex_p67(string hex)
        {
            hex = hex.ToUpper();
            if (hex.Contains("#")) hex = hex.Replace("#", "");
            var redHex = hex[0].ToString() + hex[1];
            var greenHex = hex[2].ToString() + hex[3];
            var blueHex = hex[4].ToString() + hex[5];
            var r = GetRGBPerHex_p67(redHex) / 255f;
            var g = GetRGBPerHex_p67(greenHex) / 255f;
            var b = GetRGBPerHex_p67(blueHex) / 255f;
            HSVColor c = HSVColorPicker.RGBToHSV(r, g, b);
            return c;
        }

        private static int GetRGBPerHex_p67(string code)
        {
            var result = 0;
            switch (code[0])
            {
                case '1':
                    result += 16;
                    break;
                case '2':
                    result += 32;
                    break;
                case '3':
                    result += 48;
                    break;
                case '4':
                    result += 64;
                    break;
                case '5':
                    result += 80;
                    break;
                case '6':
                    result += 96;
                    break;
                case '7':
                    result += 112;
                    break;
                case '8':
                    result += 128;
                    break;
                case '9':
                    result += 144;
                    break;
                case 'A':
                    result += 160;
                    break;
                case 'B':
                    result += 176;
                    break;
                case 'C':
                    result += 192;
                    break;
                case 'D':
                    result += 208;
                    break;
                case 'E':
                    result += 224;
                    break;
                case 'F':
                    result += 240;
                    break;
            }
            switch (code[0])
            {
                case '1':
                    result += 1;
                    break;
                case '2':
                    result += 2;
                    break;
                case '3':
                    result += 3;
                    break;
                case '4':
                    result += 4;
                    break;
                case '5':
                    result += 5;
                    break;
                case '6':
                    result += 6;
                    break;
                case '7':
                    result += 7;
                    break;
                case '8':
                    result += 8;
                    break;
                case '9':
                    result += 9;
                    break;
                case 'A':
                    result += 10;
                    break;
                case 'B':
                    result += 11;
                    break;
                case 'C':
                    result += 12;
                    break;
                case 'D':
                    result += 13;
                    break;
                case 'E':
                    result += 14;
                    break;
                case 'F':
                    result += 15;
                    break;
            }
            return result;
        }

        public static bool IsHair_p67(string s)
        {
            s = s.ToLower();
            if (s.Contains("lashes")) return false;
            if (s.Contains("pubic")) return false;
            return true;
        }
        public static string HtmlDecode_p67(string s)
        {
            var dictionary = new Dictionary<string, string> {
                { "&amp;","&" },
                { "&lt;","<" },
                { "&gt;",">" },
                { "&quot;","\"" },
                { "&#39;","'" },
                { "&#x2F;","/" },
                { "&#x60;","`" },
                { "&#x3D;","=" }
            };

            List<string> keys = new List<string>(dictionary.Keys);
            foreach (var k in keys)
            {
                s = s.Replace(k, dictionary[k]);
            }
            return s;
        }

        public static bool IsLoading_p67()
        {
            if (SuperController.singleton.loadingIcon.gameObject.activeSelf) return true;
            if (SuperController.singleton.loadingUI.gameObject.activeSelf) return true;
            return false;
        }
        public static bool SkipFloorAtom(Atom a)
        {
            if (a == null) return true;
            if (!a.on) return true;
            if (!a.collisionEnabled) return true;
            if (a.type == "PlayerNavigationPanel") return true;
            if (a.type == "Person") return true;
            if (a.name.Contains("SPQRFPS")) return true;
            if (a.name == "eyeTarget") return true;
            //if (a.type == "SubScene") return true;
            if (a.type == "InvisibleLight") return true;
            if (a.type == "AudioSource") return true;
            if (a.name.Contains("Alive")) return true;
            //SuperController.LogMessage("FLOOR:" + a.name);
            return false;
        }

        public static Vector3 VectorSmoothStep(Vector3 from, Vector3 to, float t)
        {
            return Vector3.Lerp(from, to, Mathf.SmoothStep(0f, 1f, t));
        }
        public static Vector3 VectorSmoothStepAngle(Vector3 from, Vector3 to, float t)
        {
            var x = Mathf.LerpAngle(from.x, to.x, Mathf.SmoothStep(0f, 1f, t));
            var y = Mathf.LerpAngle(from.y, to.y, Mathf.SmoothStep(0f, 1f, t));
            var z = Mathf.LerpAngle(from.z, to.z, Mathf.SmoothStep(0f, 1f, t));

            return new Vector3(x, y, z);
        }
        public static float NormalizeAngle_p67(float a)
        {
            a = a % 360f;
            if (a > 180f)
            {
                a = -360 + a;
            }
            return a;
        }
        public static void RBAlignAndParent_p67(string atomid, Rigidbody rb)
        {
            Atom RAE = SuperController.singleton.GetAtomByUid(atomid);
            if (RAE == null) return;
            RAE.ClearParentAtom();
            var t = rb.transform;
            FreeControllerV3 raeController = RAE.GetStorableByID("control") as FreeControllerV3;
            if (raeController == null) return;


            RAE.mainController.currentPositionState = FreeControllerV3.PositionState.On;
            RAE.mainController.currentRotationState = FreeControllerV3.RotationState.On;


            raeController.transform.position = t.position;
            raeController.transform.rotation = t.rotation;

            RAE.mainController.currentPositionState = FreeControllerV3.PositionState.ParentLink;
            RAE.mainController.currentRotationState = FreeControllerV3.RotationState.ParentLink;
            RAE.mainController.SelectLinkToRigidbody(rb, FreeControllerV3.SelectLinkState.PositionAndRotation);




        }

        public static void RBParent_p67(string atomid, Rigidbody rb)
        {
            Atom RAE = SuperController.singleton.GetAtomByUid(atomid);
            if (RAE == null) return;
            RAE.ClearParentAtom();

            FreeControllerV3 raeController = RAE.GetStorableByID("control") as FreeControllerV3;
            if (raeController == null) return;

            if (rb == null)
            {
                RAE.mainController.currentPositionState = FreeControllerV3.PositionState.On;
                RAE.mainController.currentRotationState = FreeControllerV3.RotationState.On;
            }
            else
            {
                RAE.mainController.currentPositionState = FreeControllerV3.PositionState.ParentLink;
                RAE.mainController.currentRotationState = FreeControllerV3.RotationState.ParentLink;
                RAE.mainController.SelectLinkToRigidbody(rb, FreeControllerV3.SelectLinkState.PositionAndRotation);
            }



        }



        public static Vector3 SetVectorValByString(Vector3 v, string x, float val)
        {
            switch (x)
            {
                case "x": v.x = val; break;
                case "y": v.y = val; break;
                case "z": v.z = val; break;
            }
            return v;
        }
        public static float GetVectorValByString(Vector3 v, string x)
        {
            switch (x)
            {
                case "x": return v.x;
                case "y": return v.y;
                case "z": return v.z;
            }
            return 0f;
        }
        public static string GetSerializedDictionaryKeys(List<string> keys)
        {
            string res = "";
            string sep = "";
            foreach (var itemk in keys)
            {
                res += sep + itemk;
                sep = ",";
            }
            return res;
        }
        static IEnumerator LerpPosition_p67(Transform transform, Vector3 targetPosition, float duration)
        {
            float time = 0;
            Vector3 startPosition = transform.position;
            while (time < duration)
            {
                transform.position = Vector3.Lerp(startPosition, targetPosition, time / duration);
                time += Time.deltaTime;
                yield return null;
            }
            transform.position = targetPosition;
        }

        public static string RandomFromString_p67(string[] elements)
        {
            Random2 rnd = new Random2();
            int index = rnd.Next(elements.Length);
            return elements[index];
        }
        public static Vector3 Bezier_p67(Vector3 s, Vector3 p, Vector3 e, float t)
        {
            float rt = 1 - t;
            return rt * rt * s + 2 * rt * t * p + t * t * e;
        }

        public static float Bezier_p67(float s, float p, float e, float t)
        {
            float rt = 1 - t;
            var r = rt * rt * s + 2 * rt * t * p + t * t * e;

            var min = s < e ? s : e;
            var max = s < e ? e : s;
            r = SPQRUtils.Clamp_p67(r, min, max);
            if (r > 0f && r <= 0.05f) r = 0f;
            return r;
        }
        public static void AtomKinematicsReset_p67(Atom a)
        {
            AtomKinematicsSet_p67(a, false);
            AtomKinematicsSet_p67(a, true);
        }
        public static void AtomKinematicsSet_p67(Atom a, bool on)
        {

            foreach (Rigidbody rigidbody in a.linkableRigidbodies)
            {
                if (rigidbody.name == "control") continue;
                if (!rigidbody.name.Contains("Control")) continue;
                rigidbody.isKinematic = on;
            }
        }

        public static void AtomResetVelocities_p67(Atom a)
        {
            foreach (Rigidbody rigidbody in a.rigidbodies)
            {
                rigidbody.velocity = Vector3.zero;
            }
        }
        public static Atom getAtomFromTransform_p67(Transform t)
        {
            //  SuperController.LogMessage("recurse " + Time.fixedTime);
            Atom atom = null;
            var ids = SuperController.singleton.GetAtomUIDs();
            if (t == null) return atom;
            if (ids.Contains(t.name)) return SPQRUtils.GetAtomById_p67(t.name);

            if (t.parent != null && atom == null)
            {
                atom = SPQRUtils.getAtomFromTransform_p67(t.parent.transform);
            }

            return atom;
        }
        public static float getYAngle_p67(FreeControllerV3 a, FreeControllerV3 target)
        {
            var t = a.transform.InverseTransformDirection(target.transform.position - a.transform.position);
            var angle = -1 * Vector3.SignedAngle(Vector3.ProjectOnPlane(t, Vector3.up), Vector3.forward, Vector3.up);
            return angle;
        }


        public static void AtomLockHorizontal_p67(string uid)
        {
            SPQRUtils.setAtomStorableBool_p67(uid, "control", "yPositionLock", true);
            SPQRUtils.setAtomStorableBool_p67(uid, "control", "xRotationLock", true);
            SPQRUtils.setAtomStorableBool_p67(uid, "control", "zRotationLock", true);
        }

        public static bool isBeingHeld_p67(FreeControllerV3 _fc)
        {
            if (_fc.currentPositionState == FreeControllerV3.PositionState.Hold) return true;
            if (_fc.linkToRB != null)
            {
                if (_fc.linkToRB.name == "SelectionHandles") return true;
                if (_fc.linkToRB.name == "MouseGrab") return true;
                if (_fc.linkToRB.name == "RightHandAnchor") return true;
                if (_fc.linkToRB.name == "LeftHandAnchor") return true;

            }
            return false;
        }

        public static float AngleFix2_p67(float angle)
        {
            angle %= 360f;
            if (angle < 0f) angle = 360f - angle;
            return angle;
        }

        public static float AngleFix_p67(float angle)
        {
            angle %= 360f;
            return Mathf.Repeat(angle + 180, 360) - 180;
        }
        public static Rigidbody GetAtomRigidbody_p67(Atom a, string name)
        {
            var rbody = a.rigidbodies.First(rb => rb.name == name);
            return rbody;
        }
        public static Rigidbody GetAtomRigidbody_p67(Atom a)
        {
            var rbody = a.rigidbodies.First();
            return rbody;
        }


        public static int Clamp_p67(int value, int min, int max)
        {
            return (value < min) ? min : (value > max) ? max : value;
        }

        public static float Clamp_p67(float value, float min, float max)
        {
            return (value < min) ? min : (value > max) ? max : value;
        }
        public static string formatTime_p67(float timer)
        {
            int totalseconds = (int)(timer % 60);
            int minutes = (int)(timer / 60);
            string seconds = (totalseconds % 60).ToString();
            if (seconds.Length < 2) seconds = "0" + seconds;
            return minutes + ":" + seconds;
        }

        public static string getAtomPrefix_p67(Atom a)
        {
            if (a == null) return "";
            var name = a.uid;
            var prefix = name.TrimEnd('/').Remove(name.LastIndexOf('/') + 1);
            return prefix;
        }

        public static HSVColor GetRandomColor_p67()
        {
            float r = (float)UnityEngine.Random.Range(0f, 1f);
            float g = (float)UnityEngine.Random.Range(0f, 1f);
            float b = (float)UnityEngine.Random.Range(0f, 1f);

            HSVColor c = HSVColorPicker.RGBToHSV(r, g, b);
            return c;
        }
        public static HSVColor GetRandomColorGrey_p67()
        {
            float r = (float)UnityEngine.Random.Range(0f, 1f);

            HSVColor c = HSVColorPicker.RGBToHSV(r, r, r);
            return c;
        }


        public static HSVColor GetColor_p67(float r, float g, float b)
        {
            HSVColor c = HSVColorPicker.RGBToHSV(r, g, b);
            return c;
        }
        public static string getFrontId_p67()
        {
            return "SPQRUISubscene/SPQRUISubsceneFront";
        }
        public static string getFrontId_p67(string s)
        {
            return getFrontId_p67() + "/" + s;
        }


        public static void showSPQRUILeftPanel_p67()
        {
            SPQRUtils.ShowAtom_p67("SPQRUISubscene/SPQRUILeftPanelCitizen");
            SPQRUtils.showCam_p67();

        }
        public static void hideSPQRUILeftPanel_p67()
        {
            SPQRUtils.HideAtom_p67("SPQRUISubscene/SPQRUILeftPanelCitizen");
            SPQRUtils.hideCam_p67();

        }

        public static void showSPQRUI_p67()
        {
            SPQRUtils.ShowAtom_p67("SPQRUISubscene");
        }
        public static void hideSPQRUI_p67()
        {
            SPQRUtils.HideAtom_p67("SPQRUISubscene");
            SPQRUtils.hideSPQRUILeftPanel_p67();
        }

        public static void showCam_p67()
        {
            SPQRUtils.setAtomStorableBool_p67("WindowCamera", "CameraControl", "cameraOn", true);

            // Atom Cam = SPQRUtils.GetAtomById_p67("WindowCamera");
            // if(Cam == null) return;
            // CameraControlUI CamUI = Cam.gameObject.GetComponentInChildren<CameraControlUI>(true);
            // if(CamUI == null) return;
            // CamUI.cameraOnToggle.isOn = true;
        }
        public static void hideCam_p67()
        {
            SPQRUtils.setAtomStorableBool_p67("WindowCamera", "CameraControl", "cameraOn", false);
            // Atom Cam = SPQRUtils.GetAtomById_p67("WindowCamera");
            // if(Cam == null) return;
            // CameraControlUI CamUI = Cam.gameObject.GetComponentInChildren<CameraControlUI>(true);
            // if(CamUI == null)return;
            // CamUI.cameraOnToggle.isOn = false;
        }

        public static void loadSubscene_p67(MVRScript X, string atomId, string category, string subId)
        {
            string path = SPQRUtils.Path_p67(SPQRUtils.GetPackagePath_p67(X) + "Custom/SubScene/SPQR/" + category + "/" + subId + ".json");
            //SPQRUtils.Log_p67("path:: " + path);

            // SuperController.LogMessage(atomId + " --- " + path);
            setAtomStorableUrl_p67(atomId, "SubScene", "storePath", path);
            Atom a = SPQRUtils.GetAtomById_p67(atomId);
            SubScene t = a.GetStorableByID("SubScene") as SubScene;
            if (t == null) return;
            t.LoadSubScene();

            a.SetOn(false);
            X.StartCoroutine(SPQRUtils.CoroutineSetAtomOn_p67(a, true, 1f));
        }



        public static IEnumerator CoroutineSetAtomOn_p67(Atom a, bool on, float delay)
        {
            if (a == null) yield break;
            yield return new WaitForSeconds(delay);
            a.SetOn(on);
        }

        public static void loadSubscene_p67(string atomId, string subId, string prefix)
        {
            string path = prefix + SPQRUtils.Path_p67(@"Custom/SubScene/SPQR/Aeternum/" + subId + ".json");
            //SPQRUtils.Log_p67("path:: " + path);
            setAtomStorableUrl_p67(atomId, "SubScene", "storePath", path);
            Atom a = SPQRUtils.GetAtomById_p67(atomId);
            SubScene t = a.GetStorableByID("SubScene") as SubScene;
            if (t == null) return;
            t.LoadSubScene();
            // SubSceneUI atomui = a.gameObject.GetComponentInChildren<SubSceneUI>(true);
            // if(atomui==null) return;

        }

        public static void ClearSubscene_p67(string atomId)
        {

            Atom a = SPQRUtils.GetAtomById_p67(atomId);
            if (a == null) return;
            SubScene t = a.GetStorableByID("SubScene") as SubScene;
            if (t == null) return;
            t.PreRemove();
        }


        public static void parentLinkAtPosition_p67(Atom a, Atom parent, JSONStorable toPosition, string toPositionName)
        {
            a.parentAtom = parent;
            Vector3 p = toPosition.transform.position;
            a.mainController.currentPositionState = FreeControllerV3.PositionState.On;
            a.mainController.transform.position = new Vector3(p.x, p.y, p.z);
            a.mainController.SetLinkToAtom(parent.uid);
            a.mainController.SetLinkToRigidbodyObject(toPositionName);
            a.mainController.currentPositionState = FreeControllerV3.PositionState.ParentLink;
            a.mainController.canGrabPosition = false;
            a.mainController.canGrabRotation = false;
        }


        public static Text GetAtomText_p67(string id)
        {
            Atom spqruititle = SPQRUtils.GetAtomById_p67(id);
            TextStorable spqruitiletext = spqruititle.GetStorableByID("Text") as TextStorable;
            if (spqruitiletext == null) return null;
            return spqruitiletext.displayField;
        }



        public static void Xfunction_p67(string f, string p1, string p2, string p3)
        {
            setAtomStorableString_p67("SPQRUISubscene/SPQRUI", "plugin#0_SPQR.SPQRUI", "varParam1", p1);
            setAtomStorableString_p67("SPQRUISubscene/SPQRUI", "plugin#0_SPQR.SPQRUI", "varParam2", p2);
            setAtomStorableString_p67("SPQRUISubscene/SPQRUI", "plugin#0_SPQR.SPQRUI", "varParam3", p3);
            setAtomStorableString_p67("SPQRUISubscene/SPQRUI", "plugin#0_SPQR.SPQRUI", "varFunction", f);
        }

        public static NameValueCollection variables;

        public static string Path_p67(string s)
        {
            return s;
        }

        public static void AddEvent_p67(string eventString)
        {
            Atom q = GetAtomById_p67("SPQREventQueue1");
            JSONStorable t = q.GetStorableByID("Text");
            JSONStorableString tval = t.GetStringJSONParam("text");
            tval.val = eventString;
        }

        public static string PopEvent_p67()
        {
            string s;
            Atom q = GetAtomById_p67("SPQREventQueue1");
            JSONStorable t = q.GetStorableByID("Text");
            JSONStorableString tval = t.GetStringJSONParam("text");
            s = tval.val + "";
            tval.val = "";
            return s;
        }

        // Invoke_p67("MyFunctionNoParams", 1f);
        // this.Invoke_p67(MyFunctionNoParams, 1f);
        // this.Invoke_p67(() => MyFunctionParams(0, false), 1f);

        public static void Invoke_p67(this MonoBehaviour mb, Action f, float delay)
        {
            mb.StartCoroutine(InvokeRoutine_p67(f, delay));
        }

        private static IEnumerator InvokeRoutine_p67(System.Action f, float delay)
        {
            yield return new WaitForSeconds(delay);
            f();
        }

        public static void SetAtomStorableColor_p67(string atom_id, string storable_id, string param_id, HSVColor newValue)
        {
            var q = SPQRUtils.GetAtomById_p67(atom_id);
            if (q == null) return;
            JSONStorable t = q.GetStorableByID(storable_id);
            JSONStorableColor tval = t.GetColorJSONParam(param_id);
            tval.val = newValue;
        }

        public static JSONStorableColor GetAtomStorableColor_p67(string atom_id, string storable_id, string param_id)
        {
            var q = SPQRUtils.GetAtomById_p67(atom_id);
            JSONStorable t = q.GetStorableByID(storable_id);
            JSONStorableColor tval = t.GetColorJSONParam(param_id);
            return tval;
        }

        public static void setAtomStorableBool_p67(string atom_id, string storable_id, string param_id, bool newValue)
        {
            var q = SPQRUtils.GetAtomById_p67(atom_id);
            if (q == null) return;
            JSONStorable t = q.GetStorableByID(storable_id);
            JSONStorableBool tval = t.GetBoolJSONParam(param_id);
            tval.val = newValue;
        }

        public static void SetAtomStorableFloat_p67(string atom_id, string storable_id, string param_id, float newValue)
        {
            var q = SPQRUtils.GetAtomById_p67(atom_id);
            if (q == null) return;
            JSONStorable t = q.GetStorableByID(storable_id);
            if (t == null) return;
            JSONStorableFloat tval = t.GetFloatJSONParam(param_id);
            if (tval == null) return;
            tval.val = newValue;
        }

        public static JSONStorableBool getAtomStorableBool_p67(string atom_id, string storable_id, string param_id)
        {
            var q = SPQRUtils.GetAtomById_p67(atom_id);
            JSONStorable t = q.GetStorableByID(storable_id);
            JSONStorableBool tval = t.GetBoolJSONParam(param_id);
            return tval;
        }
        public static JSONStorableFloat getAtomStorableFloat_p67(string atom_id, string storable_id, string param_id)
        {
            var q = SPQRUtils.GetAtomById_p67(atom_id);
            JSONStorable t = q.GetStorableByID(storable_id);
            JSONStorableFloat tval = t.GetFloatJSONParam(param_id);
            return tval;
        }



        public static void setAtomStorableString_p67(string atom_id, string storable_id, string param_id, string newValue)
        {
            var q = SPQRUtils.GetAtomById_p67(atom_id);
            if (q == null) return;
            JSONStorable t = q.GetStorableByID(storable_id);
            JSONStorableString tval = t.GetStringJSONParam(param_id);
            tval.val = newValue;
        }

        public static JSONStorableString getAtomStorableString_p67(string atom_id, string storable_id, string param_id)
        {
            var q = SPQRUtils.GetAtomById_p67(atom_id);
            JSONStorable t = q.GetStorableByID(storable_id);
            JSONStorableString tval = t.GetStringJSONParam(param_id);
            return tval;
        }
        public static void setAtomStorableStringChooser_p67(string atom_id, string storable_id, string param_id, string newValue)
        {
            var q = SPQRUtils.GetAtomById_p67(atom_id);
            if (q == null) return;
            JSONStorable t = q.GetStorableByID(storable_id);
            JSONStorableStringChooser tval = t.GetStringChooserJSONParam(param_id);
            tval.val = newValue;
        }

        public static JSONStorableStringChooser getAtomStorableStringChooser_p67(string atom_id, string storable_id, string param_id)
        {
            var q = SPQRUtils.GetAtomById_p67(atom_id);
            JSONStorable t = q.GetStorableByID(storable_id);
            JSONStorableStringChooser tval = t.GetStringChooserJSONParam(param_id);
            return tval;
        }

        public static void setAtomStorableUrl_p67(string atom_id, string storable_id, string param_id, string newValue)
        {
            var q = SPQRUtils.GetAtomById_p67(atom_id);
            if (q == null) return;
            JSONStorable t = q.GetStorableByID(storable_id);
            JSONStorableUrl tval = t.GetUrlJSONParam(param_id);
            tval.val = newValue;
        }

        public static JSONStorableUrl getAtomStorableUrl_p67(string atom_id, string storable_id, string param_id)
        {
            var q = SPQRUtils.GetAtomById_p67(atom_id);
            JSONStorable t = q.GetStorableByID(storable_id);
            JSONStorableUrl tval = t.GetUrlJSONParam(param_id);
            return tval;
        }

        public static Atom GetAtomById_p67(string id)
        {
            try
            {
                var a = SuperController.singleton.GetAtomByUid(id);
                if (a == null)
                {
                    //SuperController.LogError("not found : " + id);
                }
                return a;
            }
            catch (Exception e)
            {

                //  SuperController.LogError("Exception caught: " + e);
            }

            return null;
        }


        public static string GetFieldText_p67(string id)
        {
            Atom A = SPQRUtils.GetAtomById_p67(id);
            string res = "";
            if (A == null) return res;
            JSONStorable fieldFormula = A.GetStorableByID("Text");
            JSONStorableString fieldFormulaText = fieldFormula.GetStringJSONParam("text");
            return fieldFormulaText.val;
        }

        public static void SetFieldText_p67(string id, string t)
        {
            Atom A = SPQRUtils.GetAtomById_p67(id);
            if (A == null) return;
            JSONStorable fieldFormula = A.GetStorableByID("Text");
            JSONStorableString fieldFormulaText = fieldFormula.GetStringJSONParam("text");
            fieldFormulaText.val = t;
        }

        public static void Log_p67(string t)
        {
            SuperController.LogMessage(t);
        }



        public static NameValueCollection DecodeData_p67(string sData)
        {
            NameValueCollection result = new NameValueCollection();
            string[] pairs = sData.Split('&');
            foreach (var pair in pairs)
            {
                string[] pairData = pair.Split('=');
                result.Add(pairData[0], pairData[1]);
            }
            return result;
        }

        public static void HideAtom_p67(string id)
        {
            Atom atom = GetAtomById_p67(id);
            if (atom == null) return;
            atom.SetOn(false);
        }
        public static void ShowAtom_p67(string id)
        {
            Atom atom = GetAtomById_p67(id);
            if (atom == null) return;
            atom.SetOn(true);
        }

        public static string TimePretty_p67()
        {
            return DateTime.Now.ToString("MM/dd/yyyy HH:mm:ss");
        }
        public static string Timestamp_p67()
        {
            return ((Int32)(DateTime.Now.Subtract(new DateTime(1970, 1, 1))).TotalSeconds).ToString();
        }
        public static JSONNode GetStorablesFromFile_p67(string pathToJSON, string storableId = null)
        {
            bool found = false;
            JSONNode foundValue = null;
            try
            {
                string aJSON = SuperController.singleton.ReadFileIntoString(pathToJSON);
                JSONNode jSONNode = JSON.Parse(aJSON);



                for (int i = 0; i < jSONNode["storables"].Count; i++)
                {
                    if ((string)jSONNode["storables"][i]["id"] == storableId)
                    {
                        foundValue = jSONNode["storables"][i];
                        found = true;
                    }
                }
            }
            catch (Exception e)
            {
                //SuperController.LogError("Exception caught: " + e);
            }

            return foundValue;

        }

        public static Dictionary<string, string> GetJsonMapFromFile_p67(string pathToJSON)
        {
            var response = new Dictionary<string, string>() { };
            try
            {
                string aJSON = SuperController.singleton.ReadFileIntoString(pathToJSON);
                JSONNode jSONNode = JSON.Parse(aJSON);

                var jc = (JSONClass)jSONNode;

                foreach (var k in jc.Keys)
                {
                    response.Add(k, jSONNode[k]); ;
                }
            }
            catch (Exception e)
            {
                //SuperController.LogError("Exception caught: " + e);
            }

            return response;

        }


        public static NameValueCollection DecodeDataCustom_p67(string sData, int custom)
        {
            NameValueCollection result = new NameValueCollection();
            int ite = 0;
            string[] pairs = sData.Split('&');
            foreach (var pair in pairs)
            {
                string[] pairData = pair.Split(':');
                if (pairData.Length < 4 || pairData[3] == null)
                {
                    continue;
                }
                if (custom == 4)
                {
                    result.Add(pairData[0] + ":" + pairData[1] + ":" + pairData[2], ite.ToString());
                }
                else
                {
                    result.Add(ite.ToString(), pairData[custom]);
                }
                ite++;
            }
            return result;
        }
        public static string DictionaryToJson_p67(Dictionary<string, string> dict)
        {
            var entries = dict.Select(d =>
                string.Format("\"{0}\": \"{1}\"", d.Key, d.Value));
            return "{" + string.Join(",", entries.ToArray()) + "}";
        }

        public static string CreateMD5Hash_p67(string input)
        {
            // Step 1, calculate MD5 hash from input
            MD5 md5 = System.Security.Cryptography.MD5.Create();
            byte[] inputBytes = System.Text.Encoding.ASCII.GetBytes(input);
            byte[] hashBytes = md5.ComputeHash(inputBytes);

            // Step 2, convert byte array to hex string
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < hashBytes.Length; i++)
            {
                sb.Append(hashBytes[i].ToString("X2"));
            }
            return sb.ToString();
        }



        public static IEnumerable<string> Split_p67(string str, int chunkSize)
        {
            return Enumerable.Range(0, (str.Length + chunkSize - 1) / chunkSize).Select(i => str.Substring(i * chunkSize, Math.Min(str.Length - i * chunkSize, chunkSize)));
        }


        public static int ToInt_p67(string s)
        {

            if (s == "" || s == null) s = "0";
            return Int32.Parse(s);
        }

        public static int ToInt_p67(float f)
        {
            return Mathf.RoundToInt(f);
        }


        public static bool checkChance_p67(float chance)
        {
            return ((float)UnityEngine.Random.Range(0f, 100f) < chance);
        }


        public static string RandomFromDictionary_p67(Dictionary<string, float> d)
        {
            Random2 rand = new Random2();
            return d.ElementAt(rand.Next(0, d.Count)).Key;

        }



        public static string[] GetAEMorphs_p67()
        {
            string[] s = {
            "605-Anus-Shape-Adj.Top",
            "606-Anus-Horizontal-Adj.",
            "607-Anus-Vertical-Adj.",
            "608-Anus-In.Out-Adj.",
            "609-Anus-In.Out-SmallArea",
            "610-Anus-In.Out-WideArea",
            "612-Anus-OuterArea-Relax",
            "621-Anus-Rotation-a",
            "622-Anus-Rotation-b",
            "623-Anus-Rotation-c",
            "626-Anus-Upper-Adj.c",
            "627-Anus-Upper-Adj.d",
            "641-Anus-OuterLine-Adj.a",
            "642-Anus-OuterLine-Adj.b",
            "Forearms Size",
            "Shoulder Width",
            "Shoulder Width (B)",
            "Shoulders Size",
            "Upper Arms Size",
            "Wrist Size",
            "Back Definition .",
            "BackLessCurve",
            "Lats Size",
            "Lower Back Dimple",
            "Sacral Dimples",
            "(REN) Thin",
            "Aiko 6 Body",
            "Androgynous",
            "Beau Body",
            "Body Scale",
            "Body Size",
            "Body Tone",
            "Bodybuilder Details",
            "Bodybuilder Size",
            "Carmen Body",
            "Daisy Body",
            "Daisy Body (REN)",
            "Elise Body (REN)",
            "Emaciated",
            "Fitness Details",
            "Fitness Size",
            "Gia Body",
            "Heather Body",
            "Heavy",
            "Height",
            "Holly Body",
            "Iris Body",
            "Jade Body (REN)",
            "Kori Body",
            "Lilith Body",
            "Lily Body",
            "Lorraine Body",
            "Lower Body Length",
            "Lucille Body",
            "Lyric Body",
            "Maria Body (VAM)",
            "Mei Lin 6 Body",
            "Monique 6 Body",
            "Olympia Body",
            "Pear Figure",
            "Poppy Body",
            "Rose Body",
            "Stephanie 6 Body",
            "Teen Josie Body",
            "Thin",
            "Victoria 4 Body",
            "Victoria 6 Body",
            "Violet Body",
            "Voluptuous",
            "Areola Depth",
            "Areola Puffy Edge",
            "Areola Size",
            "Areola Size X",
            "Areola Size Y",
            "Areolae Depth",
            "Areolae Diameter",
            "Areolae Perk",
            "Breast Centered",
            "Breast Diameter",
            "Breast Height",
            "Breast Height Lower",
            "Breast Height Upper",
            "Breast Large",
            "Breast Pointed",
            "Breast Round",
            "Breast Sag1",
            "Breast Sag2",
            "Breast Side Smoother",
            "Breast Small",
            "Breast Top Curve1",
            "Breast Top Curve2",
            "Breast Under Smoother1",
            "Breast Under Smoother2",
            "Breast Under Smoother3",
            "Breast Under Smoother4",
            "Breast Under Smoother5",
            "Breast Zero",
            "Breasts Cleavage",
            "Breasts Diameter",
            "Breasts Downward Slope",
            "Breasts Gone",
            "Breasts Heavy",
            "Breasts Implants",
            "Breasts Implants Left",
            "Breasts Implants Right",
            "Breasts Natural",
            "Breasts Natural Left",
            "Breasts Natural Right",
            "Breasts Perk Side",
            "Breasts Shape 01",
            "Breasts Shape 02",
            "Breasts Shape 03",
            "Breasts Shape 04",
            "Breasts Shape 05",
            "Breasts Shape 06",
            "Breasts Shape 07",
            "Breasts Shape 08",
            "Breasts Size",
            "Breasts Small",
            "Breasts Under Curve",
            "Breasts Upward Slope",
            "BreastsCrease",
            "BreastsFlatten",
            "BreastsShape1",
            "BreastsShape2",
            "BreastsShape3",
            "Nipple Diameter",
            "Nipple Length",
            "Nipple Size",
            "Nipples",
            "Nipples Apply",
            "Nipples Depth",
            "Nipples Diameter",
            "Nipples Large",
            "Nipples Size",
            "Armpit Curve",
            "Center Gap Depth",
            "Center Gap Height",
            "Center Gap NarrowWide",
            "Center Gap Smooth",
            "Center Gap UpDown",
            "Centre Gap Narrow",
            "Centre Gap Wide",
            "Chest Height",
            "Chest Smoother",
            "ChestSeparateBreasts",
            "ChestShape",
            "ChestSmoothCenter",
            "ChestUnderBreast",
            "ChestUp",
            "ChestUpperNarrow",
            "Collarbone Detail",
            "CollarBoneDefine",
            "CollarBoneIn",
            "CollarBoneStraight",
            "CollarNotchStrength",
            "Costal Angle Arched",
            "Costal Angle Pointed",
            "Ribcage Size",
            "RibCageDefine",
            "RibCageWidth",
            "Ribs Definition",
            "Ribs Width",
            "Scapula Depth",
            "Scapula Size",
            "Sternum Depth",
            "Sternum Height",
            "Sternum Width",
            "Feet Arch",
            "Feet Thickness",
            "Foot Length",
            "Foot Length Left",
            "Foot Length Right",
            "101-Cervix-Vagina-Inner-Base",
            "102-Cervix-Open.Close",
            "103-Vag.Tube-Back-Adj.",
            "104-Vag.Tube-Front-Adj.",
            "201-Clit-Hood-Shorter",
            "202-Clit-Hood-In.Out",
            "203-Clit-Diameter",
            "204-Clit-In.Out",
            "205-Clit-Tip-Fwd.Bkwd",
            "206-ClitTo-MonsPubisArea-Adj.",
            "631-Anus-Vagina-Bridge-Adj.a",
            "632-Anus-Vagina-Bridge-Adj.b",
            "633-Anus-Vagina-Bridge-Adj.c",
            "634-Anus-Vagina-Bridge-Adj.d",
            "635-Anus-Vagina-Bridge-Adj.e",
            "636-Anus-Vagina-Bridge-Adj.f",
            "637-Anus-Vagina-Bridge-Adj.g",
            "702-Genital-Area-Rotation",
            "703-Genital-Area-Move-In.Out",
            "704-MonsPubis-In.Out-a",
            "Clitoris-erection",
            "Clitoris-exposure",
            "Clitoris-size",
            "Gen_Innie",
            "Genitals-extreme expansion",
            "Labia majora-relaxation",
            "Labia majora-spread-LLow",
            "Labia majora-spread-LMid",
            "Labia majora-spread-LUp",
            "Labia majora-spread-RLow",
            "Labia majora-spread-RMid",
            "Labia majora-spread-RUp",
            "Labia minora-exstrophy",
            "Labia minora-relaxation",
            "Labia minora-size",
            "Labia minora-spread-LLow",
            "Labia minora-spread-LMid",
            "Labia minora-spread-LUp",
            "Labia minora-spread-RLow",
            "Labia minora-spread-RMid",
            "Labia minora-spread-RUp",
            "Labia minora-style1",
            "Labia minora-style2",
            "Labia minora-style3",
            "Labia minora-style4",
            "Labia minora-thickness",
            "Vagina-closing",
            "Vagina-expansion",
            "Claws2",
            "Hands Thickness",
            "Lorraine Nails",
            "Nails Length",
            "Brow Define",
            "Brow Define Left",
            "Brow Define Right",
            "Brow Depth",
            "Brow Depth Left",
            "Brow Depth Right",
            "Brow Height",
            "Brow Inner Height",
            "Brow Inner Width",
            "Brow Outer Depth",
            "Brow Outer Height",
            "Brow Outer Shift",
            "Brow Outer Width",
            "Brow Shape Inner",
            "Brow Shape Middle",
            "Brow Shape Outer",
            "Brow Width",
            "Brow Width Left",
            "Brow Width Right",
            "BrowAreaSmooth",
            "BrowAreaSmooth2",
            "Brows Arch",
            "Brows Size",
            "Cheek Bone Define ",
            "Cheek Bones Arch ",
            "Cheek Bones Height",
            "Cheek Bones Low",
            "Cheek Bones Round",
            "Cheek Bones Shape 1",
            "Cheek Bones Size",
            "Cheek Bones Size Left",
            "Cheek Bones Size Right",
            "Cheek Bones Upper Depth",
            "Cheek Bones Width",
            "Cheek Bones Width Middle ",
            "Cheek Bones Width Outer",
            "Cheek Bones Width Upper",
            "Cheek Jowl",
            "Cheek Lower Depth",
            "Cheek Lower Width",
            "Cheek Squish Left",
            "Cheek Squish Right",
            "CheekBones1",
            "CheekBonesDefine",
            "CheekBonesHigh",
            "CheekBonesLarge",
            "CheekBonesMore",
            "CheekHollows",
            "CheekInnerPuff",
            "Cheeks Define",
            "Cheeks Depth",
            "Cheeks Depth Left",
            "Cheeks Depth Middle",
            "Cheeks Depth Right",
            "Cheeks Depth Upper",
            "Cheeks Dimple Crease",
            "Cheeks Dimple Crease Left",
            "Cheeks Dimple Crease Right",
            "Cheeks Flat",
            "Cheeks Height",
            "Cheeks Inner Height",
            "Cheeks Inner Puffy",
            "Cheeks Puffy Lower ",
            "Cheeks Sink",
            "Cheeks Sink Left",
            "Cheeks Sink Lower",
            "Cheeks Sink Right",
            "Cheeks Upper Crease",
            "CheekShape3",
            "CheekShapeRounder",
            "CheekShapeRounder2",
            "CheeksOut",
            "CheeksPuffLower",
            "CheeksShape1",
            "CheeksSmooth",
            "Jennifer Cheeks",
            "Laugh Lines",
            "Chin Cleft",
            "Chin Crease",
            "Chin Crease B",
            "Chin Crease Smooth",
            "Chin Depth",
            "Chin Height",
            "Chin Round",
            "Chin Square",
            "Chin Width",
            "Chin Width 2",
            "Chin Width Left",
            "Chin Width Right",
            "ChinCleft",
            "ChinOut",
            "ChinOutRound",
            "ChinShape",
            "ChinWider",
            "Earlobes Attached",
            "Earlobes Length",
            "Earlobes Size",
            "Ears  Angle Upper",
            "Ears Angle",
            "Ears Angle Left",
            "Ears Angle Right",
            "Ears Depth",
			// "Ears Elf",
			// "Ears Elf Long",
			"Ears Height",
            "Ears Height Left",
            "Ears Height Right",
            "Ears Size",
            "Ears Size Left",
            "Ears Size Right",
			// "FairyEars1",
			// "FairyEars2",
			// "FairyEars3",
			"Cornea Depth",
            "Crows Feet",
            "Eye Fold",
            "Eye Shape Bottom M",
            "Eye Upper Shape M",
            "EyeAreaSmooth",
            "EyeAreaThin",
            "Eyeballs Depth",
            "Eyelashes Curl",
			// "Eyelashes Hide Layer 1",
			// "Eyelashes Hide Layer 2",
			"Eyelashes Length",
            "Eyelashes Top Point",
            "Eyelid Lower Crease",
            "Eyelid Lower Shape",
            "Eyelid Upper Inner Shape",
            "Eyelid Upper Shape  Outer",
            "Eyelids Bottom Define",
            "Eyelids Bottom In Height",
            "Eyelids Bottom Out Height",
            "Eyelids Fold Down",
            "Eyelids Heavy",
            "Eyelids Height Inner ",
            "Eyelids Lower Height",
            "Eyelids Lower Puffy",
            "Eyelids Shape 1",
            "Eyelids Shape 2",
            "Eyelids Smooth",
            "Eyelids Top In Height",
            "Eyelids Top Out Height",
            "Eyelids Upper Height",
            "Eyelids Upper Height (Outer)",
            "Eyelids Upper Height A",
            "Eyelids Upper Height Left",
            "Eyelids Upper Height Right",
            "Eyelids Upper Square",
            "EyeOutUpDown",
            "Eyes Almond Inner",
            "Eyes Almond Inner Left",
            "Eyes Almond Inner Right",
            "Eyes Almond Outer",
            "Eyes Almond Outer Left",
            "Eyes Almond Outer Right",
            "Eyes Angle",
            "Eyes Angle Left",
            "Eyes Angle Right",
            "Eyes Bags",
            "Eyes Closed (REN)",
            "Eyes Cornea Bulge",
            "Eyes Depth",
            "Eyes Depth Left",
            "Eyes Depth Right",
            "Eyes Height",
            "Eyes Height Bottom",
            "Eyes Height Inner",
            "Eyes Height Left",
            "Eyes Height Outer",
            "Eyes Height Right",
            "Eyes Height Upper",
            "Eyes Inner  Shape",
            "Eyes Inner Corner Height",
            "Eyes Inner Corner Width",
            "Eyes Inner Depth",
            "Eyes Inner Square",
            "Eyes Iris Correction",
            "Eyes Iris Size",
            "Eyes Lower Shape",
            "Eyes Open",
            "Eyes Outer Shape",
            "Eyes Puffy Lower",
            "Eyes Puffy Outer",
            "Eyes Puffy Shift",
            "Eyes Puffy Upper",
            "Eyes Puffy Upper Center ",
            "Eyes Round",
            "Eyes Round Lower",
            "Eyes Round Upper",
            "Eyes Shift Lower",
            "Eyes Shift Upper",
			// "Eyes Size",
			// "Eyes Size Left",
			// "Eyes Size Right",
			"Eyes Slant Outer ",
            "Eyes Upper Shape",
            "Eyes Width",
            "Eyes Width Left",
            "Eyes Width Right",
            "Eyes Wrinkle",
            "EyeShape5",
            "EyeShape6",
            "EyeShapeLarger",
            // "EyesShape1",
            "EyesShape2",
            "EyesShape3",
            "EyesShape4",
            "Iris Height",
            "Iris Placement",
            "Iris Placement L",
            "Lacrimals Pinch",
            "Lacrimals Size",
            "LacrimalStretch",
            "LacrimalYoung",
            "Pupils Dialate",
            "Pupils Slit",
            "Teen Josie Lashes",
            "Face Angle",
            "Face Center Depth",
            "Face Depth",
            "Face Depth Lower ",
            "Face Flat",
            "Face Flat Left",
            "Face Flat Right",
            "Face Heart",
            "Face Heart Left",
            "Face Heart Right",
            "Face Height",
            "Face Height 2",
            "Face Height Lower",
            "Face Height Upper ",
            "Face Lower Thin",
            "Face Middle Thin",
            "Face Round",
            "Face Round Left",
            "Face Round Right",
            "Face Size",
            "Face Square",
            "Face Square Left",
            "Face Square Right",
            "Face Upper Width",
			// "Face Young",
			// "FaceAllWider",
			// "FaceExotic",
			// "FaceFae1",
			// "FaceFae2",
			"FaceFemale4",
            "FaceMidProportion",
            "FaceMidProportion2",
            "FaceMidProportion3",
			// "FaceModel",
			// "FaceModel2",
			"FaceRoundFlat1",
            "FaceRoundFlat2",
            "FaceShape1",
            // "FaceShape2",
            "FaceShape3",
            "FaceShapeAfrican",
            //"FaceShapeAnime",
            // "FaceShapeFemale",
			// "FaceShapeFlat1",
			// "FaceShapeFlat2",
			// "FaceShapeHeart",
            // "FaceShapeHeart2",
			// "FaceShapeOval",
			"FaceShapeRound",
            "FaceShapeRound2",
            "FaceShapeSquare",
            "FaceShapeThin",
            "FaceSmoothAll",
            "FaceSmoothLower",
            "FemaleProportion1",
            "FemaleProportion2",
            "FemaleProportion4",
            "LowerFaceSmoothRound",
            "Young1",
            "Young2",
            "Young3",
            "Young4",
            "Young5",
            "Jaw Angle",
            "Jaw Chin Shape",
            "Jaw Corner Height",
            "Jaw Corner Shape",
            "Jaw Corner Width",
            "Jaw Corner Width Left",
            "Jaw Corner Width Right",
            "Jaw Curve",
            "Jaw Define",
            "Jaw Height",
            "Jaw Height Left",
            "Jaw Height Right",
            "Jaw Line Depth",
            "Jaw Round",
            "Jaw Size",
			// "Jaw Size Left",
			// "Jaw Size Right",
			"Jaw Square",
            "JawCornerForward",
            "JawEarForward",
            "JawHeart",
            "JawlineRecede",
            "JawlineReduce",
			// "Teeth Bottom Size",
			// "Teeth Bucked",
			// "Teeth Gap",
			// "Teeth Irregular",
			// "Teeth Lower Jaw Move Forward-Back",
			// "Teeth Lower Jaw Move Side-Side",
			// "Teeth Lower Jaw Move Up-Down",
			// "Teeth Lower Jaw Size",
			// "Teeth Lower Jaw Size Depth",
			// "Teeth Lower Jaw Size Height",
			// "Teeth Lower Jaw Size Width",
			// "Teeth Top Size",
			// "Teeth Upper Jaw Move Forward-Back",
			// "Teeth Upper Jaw Move Side-Side",
			// "Teeth Upper Jaw Move Up-Down",
			// "Teeth Upper Jaw Size",
			// "Teeth Upper Jaw Size Depth",
			// "Teeth Upper Jaw Size Height",
			// "Teeth Upper Jaw Size Width",
			// "Vampire Fangs",
			// "Tongue Forked",
			// "Tongue Narrow-Wide",
			// "Tongue Ridge Bottom",
			// "Tongue Thickness",
			// "Tongue Tip Thickness",
			// "CurvyLips",
			"Lip Lower Depth",
            "Lip Lower Size",
            "Lip Lower Width",
            "Lip Top Peak",
            "Lip Upper (ren)",
            "Lip Upper Curves",
            "Lip Upper Depth",
            "Lip Upper Size",
            "Lip Upper Thick",
            "LipLoPuffy",
            "Lips Bottom Full",
            "Lips Bottom Shape ",
            "Lips Bottom Small",
            "Lips Bow Height",
            "Lips Bow Shape",
            "Lips Center Angle",
            "Lips Corners Pinch",
            "Lips Depth",
            "Lips Edge Define",
            "Lips Heart",
			// "Lips Pucker (REN)",
			"Lips Square",
            "Lips Thin",
            "Lips Top Full",
            "LIps Top Width",
            "Lips Upper Bow",
            "Lips Upper Center Depth",
            "Lips Upper Curves  Round",
            "Lips Upper Curves Corner",
            "LipsPuffOut",
            "LipUpPuffy",
            "Lower Mouth Puffy",
            "Mouth Corner Depth",
            "Mouth Corner Height",
            "Mouth Corner Width",
            "Mouth Curves",
            "Mouth Curves  Arch",
            "Mouth Curves Arch Corner",
            "Mouth Curves Center Width",
            "Mouth Curves Corner",
            "Mouth Depth",
            "Mouth Height",
            "Mouth Marionette lines",
            "Mouth Mousey",
            "Mouth Side Crease B",
            "Mouth Size",
            "Mouth Width",
            "Mouth_UpLipDefine",
            "MouthArea_InOut",
            "MouthCenterXScale",
            "MouthCornerShape",
            "MouthCurves1",
            "MouthCurves2",
            "MouthLarger1",
            "MouthLarger2",
            "MouthLarger3",
            "MouthLowLipDefine",
            "MouthLowLipLg",
            "MouthLowLipLg2",
            "MouthLowLipShape",
            "MouthOuterXScale",
            "MouthPouty",
            "MouthProportion",
            "MouthRotateUp",
			// "MouthShape1",
			// "MouthShape10",
			// "MouthShape11",
			// "MouthShape12",
			// "MouthShape2",
			// "MouthShape3",
			// "MouthShape4",
			// "MouthShape5",
			// "MouthShape6",
			// "MouthShape7",
			// "MouthShape8",
			// "MouthShape9",
			"MouthSmaller1",
            "Outer Mouth Shape V",
            "SmileMuscles",
            "UpLipAreaShape",
            "Upper Lip Curve",
            "Upper Lip Puffy",
            "Uvula Size",
            "EyesNoseWidth",
            "Mack Nose",
            "NarrowNostrils",
            "Nose A",
            "Nose Bridge Curve",
            "Nose Bridge Depth",
            "Nose Bridge Height",
            "Nose Bridge Height 2",
            "Nose Bridge Lower Depth",
            "Nose Bridge Lower Width ",
            "Nose Bridge Middle Width",
            "Nose Bridge Middlle Depth",
            "Nose Bridge Root Width",
            "Nose Bridge Skew",
            "Nose Bridge Slope",
            "Nose Bridge Width",
            "Nose Bump",
            "Nose Depth",
            "Nose Flesh Size",
            "Nose Flesh Size Left",
            "Nose Flesh Size Right",
            "Nose G",
            "Nose Height",
            // "Nose Pinch",
            "Nose Ridge",
            "Nose Ridge Width",
            "Nose Septum Depth",
            "Nose Septum Height",
            "Nose Septum Shape",
            "Nose Septum Width",
            "Nose Side-Side",
            "Nose Size",
            "Nose Skew",
            "Nose Tilt",
            "Nose Tip Bottom Shape",
            "Nose Tip Depth",
            "Nose Tip Height",
            "Nose Tip Round",
            "Nose Tip Square",
            "Nose Tip Square Bottom",
            "Nose Tip Upper Shape ",
            "Nose Tip Width",
            "Nose Tip Width Lower",
            "Nose Tip Width Upper",
            "Nose Twist",
            "Nose V",
            "Nose Width",
            "NoseAreaSmooth",
            "NoseBeak",
            // "NoseFlat",
            // "NoseFlat2",
            // "NoseFlat3",
            // "NoseFlatter",
            // "NoseMouthRatio",
            "NoseRidgeReduce",
            // "NoseShape1",
            // "NoseShape10",
            // "NoseShape11",
            // "NoseShape12",
            // "NoseShape2",
            // "NoseShape3",
            // "NoseShape4",
            // "NoseShape5",
            // "NoseShape6",
            // "NoseShape7",
            // "NoseShape8",
            // "NoseShape9",
            "NoseTip1",
            "NoseTip2",
            "NoseTurnedUp",
            "Nostrils Bottom Rotate",
            "Nostrils Bottom Shape 2",
            "Nostrils Define Upper",
            "Nostrils Depth",
            "Nostrils Flare",
            "Nostrils Flesh Size",
            "Nostrils Height",
            "Nostrils Height Left",
            "Nostrils Height Right",
            "Nostrils Inner Height",
            "Nostrils Inner Width",
            "Nostrils Rotation",
            "Nostrils Shape Bottom",
            "Nostrils Shape Height",
            "Nostrils Shape Middle ",
            "Nostrils Shape Top",
            "Nostrils Shape Top Angle",
            "Nostrils Thin",
            "Nostrils Width",
            "Nostrils Width Left",
            "Nostrils Width Lower",
            "Nostrils Width Right",
            "Philtrum Angle",
            "Philtrum Depth",
            "Philtrum Width",
            "Septum Width Upper",
            "Cranium Height",
            "Cranium Size",
            "Cranium Slope",
            "Cranium Width",
            "Forehead Define",
            "Forehead Flat",
            "Forehead Height",
            "Forehead Round",
            "Forehead Slope",
            "ForeHead Top Width",
            "Forehead Width",
            "Forehead Wrinkle",
            "Forehead Wrinkle Left",
            "Forehead Wrinkle Right",
            "Head Length",
            "Head Scale",
            "Head Width",
            "Temples Define",
            "Temples Define Left",
            "Temples Define Right",
            "Adrianna",
            "Aiko 6 Head",
            "Ana Sculpt 2",
            "Aneta",
            "Beau Head",
            "Candy",
            "Carmen Face",
            "Daisy (REN)",
            "Danika Head",
            "Destiny Head",
            "Eisa",
            "Gia Head",
            "Greta",
            "Izarra Head",
            "Kimi (REN)",
            "Kori Head",
            "Lexi (REN)",
            "Lilith Head",
            "Lorraine Head",
            "Lucille Head",
            "Lyric Head",
            "MadelineHead",
            "Maria (VAM)",
            "Mei Lin 6 Head",
            "Mia (VAM)",
            "Monique 6 Head",
            "Neomi",
            "Olympia Head",
            "Parisa",
            "SC_Alessa",
            "SC_AlessaEars",
            "SC_Norma",
            "SI_Nyssa",
            "Stephanie 6 Head",
            "Sumiko Head",
            // "Tara Face",
            // "Tara Face B",
            // "Teen Josie Head",
            "VianneHead",
            "Victoria 4 Head",
            "Victoria 6 Head",
            "Glute Crease",
            "Glute Crease Left",
            "Glute Crease Right",
            "Glutes Lower Depth",
            "Glutes Lower Width",
            "Glutes Round",
            "Glutes Size",
            "Glutes Upper Depth",
            "Glutes Width",
            "GlutesBig",
            "GluteShape1",
            "GluteShape2",
            "GluteShape3",
            "GlutesSmall",
            "GlutesUp",
            "Hip Bone Crest",
            "Hip Bone Size",
            "Hip Size",
            "Hips High",
            "Hips Round",
            "HipShape1",
            "HipSmall",
            "HipSmall2",
            "HipSmooth",
            "HipSoftenCurves",
            "HipThighYoung",
            "Iliac Line",
            "Pubic Area Size",
            "Pubic Area Width",
            "PubicAreaOut",
            "Calves Size",
            "Knee Bones Size",
            "Legs Length",
            "LegShape1",
            "LegShape2",
            "Shins Size",
            "ShinsStraighten",
            "Thigh Bone Size",
            "Thigh Upper Gap",
            "Thighs Size",
            "Thighs Tone",
            "ThighShape",
            "ThighStraight",
            "Neck Length",
            "Neck Size",
            "Neck Width",
            "NeckBackIn",
            "NeckDefine1",
            "NeckFemaleCombo",
            "NeckShape1",
            "NeckThin",
            "ThroatBaseIn",
            "Traps Size",
            "Torso Fitness",
            "Upper Body Length",
            "Upper Torso Length",
            "AbdomenCreaseSmooth",
            "AbdomenSmooth",
            "Abs Definition",
            "Abs Muscle",
            "AbsUpperLess",
            "BellyShape1",
            "BellyShape2",
            "BellyShape3",
            "BellySideSmooth",
            "Linea Alba Depth",
            "Love Handles",
            "Navel",
            "Navel Depth",
            "Navel Hollow",
            "Navel Horizontal",
            "Navel Out",
            "Navel Size",
            "Navel Vertical",
            "Pregnant",
            "Rectus Outer Detail",
            "Rectus Width",
            "Stomach Depth",
            "Stomach Lower Depth",
            "Stomach Shape 1",
            "Stomach Soften",
            "Waist Narrow Lower",
            "Waist Narrow Upper",
            "Waist Width",
            "WaistBellyInOut",
            "WaistHipDefine",
            "WaistHourglass",
            "WaistLower",
            "WaistNarrow",
            "WaistNarrow2"


        };
            return s;
        }

        public static string[] GetAEMorphsShemale_p67()
        {
            string[] s = {
            "Genital Ribbed",
            "Jake Genital (JaR)",
            "Jimpei Genital (JaR)",
            "Johnny Genital (JaR)",
            "Julian Genital (JaR)",
            "Lee Genital (JaR)",
            "Seth Genital (JaR)",
            "Shane Genital (JaR)",
            "Frsk_Long_Point",
            "Frsk_Scale_Point",
            "Frsk_Up_Point",
            "PBMFS01",
            "PBMFS02",
            "PBMFS03",
            "PBMFS04",
            "PBMFS05",
            "Defined Head",
            "Foreskin Fold Adjust",
            "Gl_Big",
            "Gl_Height",
            "Gl_Incline",
            "Gl_Length",
            "Gl_Rim",
            "Gl_Short",
            "Gl_Slope",
            "Gl_Small",
            "Gl_Urethra_Open",
            "Gl_Urethra_Scale",
            "Gl_Width",
            "Glans_Fan",
            "Glans_Inflate",
            "Glans_Tall",
            "Glans_Wide",
            "Rim Thickness",
            "large nuts",
            "Sc_Castrate",
            "Sc_Left_Right",
            "Testicle Height Left",
            "Testicle Height Right",
            "Curve Left_Right",
            "Curve Up_Down",
            "Penis Length",
            "Penis Width",
            "Curved",
            "Sh_C_Spongiosum",
            "Sh_Girth",
            "Sh_Girth_Base",
            "Sh_Height",
            "Sh_Irregular",
            "Shaft Wrinkles",
            "Sh_Taper",
            "Shaft_Balloon",
            "Shaft Base Thick",
            "Shaft_CurveDown-Up",
            "Shaft_CurveLeft-Right",
            "Shaft_Spongiosum",
            "Shaft_Spongiosum-Wide",

        };
            return s;
            //return Merge_p67(s,GetAEMorphs_p67());
        }

        public static string[] Merge_p67(string[] first, string[] second)
        {
            var z = new string[first.Length + second.Length];
            first.CopyTo(z, 0);
            second.CopyTo(z, first.Length);
            return z;
        }


        public static string[] GetAEMorphsMale_p67()
        {
            string[] s = {


            "Darius 6 Head",
            "Gianni 6 Head",
            "Hector Head",
            "Julian Head (JaR)",
            "Lee 6 Head",
            "Michael 6 Head",
            "Ryze",
            "Scott 6 Head",

            "Darius 6 Body",
            "Gianni 6 Body",
            "Julian Body (JaR)",
            "Lee 6 Body",
            "Michael 6 Body",

            /// MALE

            "Genital Ribbed",
            "Jake Genital (JaR)",
            "Jimpei Genital (JaR)",
            "Johnny Genital (JaR)",
            "Julian Genital (JaR)",
            "Lee Genital (JaR)",
            "Seth Genital (JaR)",
            "Shane Genital (JaR)",
            "Frsk_Long_Point",
            "Frsk_Scale_Point",
            "Frsk_Up_Point",
            "PBMFS01",
            "PBMFS02",
            "PBMFS03",
            "PBMFS04",
            "PBMFS05",
            "Defined Head",
            "Foreskin Fold Adjust",
            "Gl_Big",
            "Gl_Height",
            "Gl_Incline",
            "Gl_Length",
            "Gl_Rim",
            "Gl_Short",
            "Gl_Slope",
            "Gl_Small",
            "Gl_Urethra_Open",
            "Gl_Urethra_Scale",
            "Gl_Width",
            "Glans_Fan",
            "Glans_Inflate",
            "Glans_Tall",
            "Glans_Wide",
            "Rim Thickness",
            "large nuts",
            "Sc_Castrate",
            "Sc_Left_Right",
            "Testicle Height Left",
            "Testicle Height Right",
            "Curve Left_Right",
            "Curve Up_Down",
            "Penis Length",
            "Penis Width",
            "Curved",
            "Sh_C_Spongiosum",
            "Sh_Girth",
            "Sh_Girth_Base",
            "Sh_Height",
            "Sh_Irregular",
            "Shaft Wrinkles",
            "Sh_Taper",
            "Shaft_Balloon",
            "Shaft Base Thick",
            "Shaft_CurveDown-Up",
            "Shaft_CurveLeft-Right",
            "Shaft_Spongiosum",
            "Shaft_Spongiosum-Wide",

        };
            return s;
        }



        public static UIDynamicTextField[] CreateTextInput_p67(MVRScript X, JSONStorableString labeljss, JSONStorableString s, bool rightSide = false)
        {
            var response = new UIDynamicTextField[2];
            UIDynamicTextField dtext;
            if (labeljss != null)
            {
                dtext = X.CreateTextField(labeljss, rightSide);
                dtext.height = 5f;
                dtext.backgroundColor = Color.clear;
                response[1] = dtext;
            }

            var textfield = X.CreateTextField(s, rightSide);
            textfield.backgroundColor = Color.white;
            var input = textfield.gameObject.AddComponent<InputField>();
            input.textComponent = textfield.UItext;
            s.inputField = input;


            response[0] = textfield;

            return response;
        }

        public static void RemoveTextInput_p67(MVRScript X, JSONStorableString labeljss, JSONStorableString s)
        {
            X.RemoveTextField(labeljss);
            X.RemoveTextField(s);
        }


        public static UIDynamicButton CreateButton_p67(MVRScript X, string label, UnityAction callback, bool rightSide = false)
        {
            UIDynamicButton button = X.CreateButton(label, rightSide);
            button.button.onClick.AddListener(callback);
            return button;
        }


        public static UIDynamic CreateSpacer_p67(MVRScript X, float height, bool rightSide)
        {
            UIDynamic spacer = X.CreateSpacer(rightSide);
            spacer.height = height;
            return spacer;
        }


        // package:/ or empty
        public static string GetPackagePath_p67(MVRScript X)
        {
            string id = X.name.Substring(0, X.name.IndexOf('_'));
            string filename = X.manager.GetJSON()["plugins"][id].Value;
            int index = filename.IndexOf(":/");
            if (index >= 0)
                return filename.Substring(0, index + 2);
            return string.Empty;
        }

        //full path
        public static string GetPluginPath_p67(MVRScript self)
        {
            string id = self.name.Substring(0, self.name.IndexOf('_'));
            string filename = self.manager.GetJSON()["plugins"][id].Value;
            return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
        }



    }



    // ===========================================================================================

    // TriggerHandler implementation for easier handling of custom triggers.
    // Essentially call this in your plugin init code:
    //     StartCoroutine(SimpleTriggerHandler.LoadAssets_p67());
    //
    // Credit to AcidBubbles for figuring out how to do custom triggers.
    public class SimpleTriggerHandler : TriggerHandler
    {
        public static bool Loaded { get; private set; }

        private static SimpleTriggerHandler myInstance_p67;

        private RectTransform myTriggerActionsPrefab_p67;
        private RectTransform myTriggerActionMiniPrefab_p67;
        private RectTransform myTriggerActionDiscretePrefab_p67;
        private RectTransform myTriggerActionTransitionPrefab_p67;

        public static SimpleTriggerHandler Instance
        {
            get
            {
                if (myInstance_p67 == null)
                    myInstance_p67 = new SimpleTriggerHandler();
                return myInstance_p67;
            }
        }

        public static void LoadAssets_p67()
        {
            SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal_p67());
        }

        private IEnumerator LoadAssetsInternal_p67()
        {
            foreach (var x in LoadAsset_p67("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab_p67 = p))
                yield return x;
            foreach (var x in LoadAsset_p67("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab_p67 = p))
                yield return x;
            foreach (var x in LoadAsset_p67("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab_p67 = p))
                yield return x;
            foreach (var x in LoadAsset_p67("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab_p67 = p))
                yield return x;

            Loaded = true;
        }

        private IEnumerable LoadAsset_p67(string assetBundleName, string assetName, Action<RectTransform> assign)
        {
            AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
            if (request == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
            yield return request;
            GameObject go = request.GetAsset<GameObject>();
            if (go == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
            RectTransform prefab = go.GetComponent<RectTransform>();
            if (prefab == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
            assign(prefab);
        }


        void TriggerHandler.RemoveTrigger(Trigger t)
        {
            // nothing to do
        }

        void TriggerHandler.DuplicateTrigger(Trigger t)
        {
            throw new NotImplementedException();
        }

        RectTransform TriggerHandler.CreateTriggerActionsUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionsPrefab_p67);
        }

        RectTransform TriggerHandler.CreateTriggerActionMiniUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab_p67);
        }

        RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab_p67);
        }

        RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
        {
            RectTransform rt = UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab_p67);
            rt.GetComponent<TriggerActionTransitionUI>().startWithCurrentValToggle.gameObject.SetActive(false);
            return rt;
        }

        void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
        {
            UnityEngine.Object.Destroy(rt?.gameObject);
        }
    }



    // Base class for easier handling of custom triggers.
    public abstract class CustomTrigger : Trigger
    {
        public string Name
        {
            get { return name; }
            set { name = value; myNeedInit = true; }
        }

        public string SecondaryName
        {
            get { return secondaryName; }
            set { secondaryName = value; myNeedInit = true; }
        }

        public MVRScript Owner
        {
            get; private set;
        }

        private string name;
        private string secondaryName;
        private bool myNeedInit = true;

        public CustomTrigger(MVRScript owner, string name, string secondary = null)
        {
            Name = name;
            SecondaryName = secondary;
            Owner = owner;
            handler = SimpleTriggerHandler.Instance;
        }

        public CustomTrigger(CustomTrigger other)
        {
            Name = other.Name;
            SecondaryName = other.SecondaryName;
            Owner = other.Owner;
            handler = SimpleTriggerHandler.Instance;

            JSONClass jc = other.GetJSON(Owner.subScenePrefix);
            base.RestoreFromJSON(jc, Owner.subScenePrefix, false);
        }

        public void OpenPanel()
        {
            if (!SimpleTriggerHandler.Loaded)
            {
                //  SuperController.LogError("CustomTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
                return;
            }

            triggerActionsParent = Owner.UITransform;
            InitTriggerUI();
            OpenTriggerActionsPanel();
            if (myNeedInit)
            {
                Transform panel = triggerActionsPanel.Find("Panel");
                panel.Find("Header Text").GetComponent<Text>().text = Name;
                Transform secondaryHeader = panel.Find("Trigger Name Text");
                secondaryHeader.gameObject.SetActive(!string.IsNullOrEmpty(SecondaryName));
                secondaryHeader.GetComponent<Text>().text = SecondaryName;

                InitPanel();
                myNeedInit = false;
            }
        }

        protected abstract void InitPanel();

        public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
        {
            if (jc.HasKey(Name))
            {
                JSONClass tc = jc[Name].AsObject;
                if (tc != null)
                    base.RestoreFromJSON(tc, subScenePrefix, isMerge);
            }
            else if (setMissingToDefault)
            {
                base.RestoreFromJSON(new JSONClass());
            }
        }
    }

    public class EventTrigger : CustomTrigger
    {
        public EventTrigger(MVRScript owner, string name, string secondary = null)
            : base(owner, name, secondary)
        {
        }

        public EventTrigger(EventTrigger other)
            : base(other)
        {
        }

        protected override void InitPanel()
        {
            Transform content = triggerActionsPanel.Find("Content");
            content.Find("Tab1/Label").GetComponent<Text>().text = "Event Actions";
            content.Find("Tab2").gameObject.SetActive(false);
            content.Find("Tab3").gameObject.SetActive(false);
        }

        public void Trigger()
        {
            active = true;
            active = false;
        }

        public void Trigger(List<TriggerActionDiscrete> actionsNeedingUpdateOut)
        {
            Trigger();
            for (int i = 0; i < discreteActionsStart.Count; ++i)
            {
                if (discreteActionsStart[i].timerActive)
                    actionsNeedingUpdateOut.Add(discreteActionsStart[i]);
            }
        }
    }



    public struct LineDrawer
    {
        private LineRenderer lineRenderer;
        private float lineSize;

        public Vector3 lineStart;
        public Vector3 lineEnd;
        public bool draw;

        public LineDrawer(float lineSize = 0.1f)
        {

            //SuperController.LogMessage("LineDrawer called!");
            draw = false;
            lineStart = new Vector3(0, -100f, 0);
            lineEnd = new Vector3(0, -100f, 0);
            GameObject lineObj = new GameObject("LineObj");
            lineRenderer = lineObj.AddComponent<LineRenderer>();
            //Particles/Additive
            lineRenderer.material = new Material(Shader.Find("Hidden/Internal-Colored"));

            this.lineSize = lineSize;
        }
        public bool IsVisible()
        {
            return lineRenderer != null;
        }

        private void init(float lineSize = 0.1f)
        {

            if (lineRenderer == null)
            {
                GameObject lineObj = new GameObject("LineObj");
                lineRenderer = lineObj.AddComponent<LineRenderer>();
                //Particles/Additive
                lineRenderer.material = new Material(Shader.Find("Hidden/Internal-Colored"));

                this.lineSize = lineSize;

                lineRenderer.SetPosition(0, new Vector3(0f, 0f, 0f));
                lineRenderer.SetPosition(1, new Vector3(0f, 0f, 0f));
            }
        }

        //Draws lines through the provided vertices
        public void DrawLineInGameView(Vector3 start, Vector3 end, Color color)
        {



            if (lineRenderer == null)
            {
                init(this.lineSize);
            }

            lineStart = start;
            lineEnd = end;


            if (draw != true) return;

            //Set color
            lineRenderer.startColor = color;
            lineRenderer.endColor = color;

            //Set width
            lineRenderer.startWidth = lineSize;
            lineRenderer.endWidth = lineSize;

            //Set line count which is 2
            lineRenderer.positionCount = 2;

            //Set the postion of both two lines
            lineRenderer.SetPosition(0, start);
            lineRenderer.SetPosition(1, end);
        }

        public void Destroy()
        {
            if (lineRenderer != null)
            {
                UnityEngine.Object.Destroy(lineRenderer.gameObject);
            }
        }


    }


}
